'''
Created on 2022-11-16

@author: wf
'''
import skg.graph


class Scholar(skg.graph.Node):
    """
    an instance of a scholar that writes papers to be an author
    """
    
    @classmethod
    def getSamples(cls):
        samples=[
            {
                "wikiDataId":"Q54303353",
                "name": "Stefan Decker",
                "gndId":"",
                "dblpId":"d/StefanDecker",
                "orcid":"0000-0001-6324-7164",
                "linkedInId":"",
                "googleScholarUser":"uhVkSswAAAAJ",
                "homepage":"http://www.stefandecker.org"
            },
            {
                "name": "Tim Berners-Lee",
                "wikiDataId": "Q80",
                "givenName": "Timothy",
                "familyName": "Berners-Lee" 
            }
        ]
        return samples
        
    
    def __init__(self):
        """
        constructor
        """


"""
Created on 2021-08-19

@author: wf
"""

import getpass
from unittest import TestCase
import time
import os


class Basetest(TestCase):
    """
    base test case
    """

    def setUp(self, debug=False, profile=True):
        """
        setUp test environment
        """
        TestCase.setUp(self)
        self.debug = debug
        self.profile = profile
        msg = f"test {self._testMethodName}, debug={self.debug}"
        self.profiler = Profiler(msg, profile=self.profile)

    def tearDown(self):
        TestCase.tearDown(self)
        self.profiler.time()

    @staticmethod
    def inPublicCI():
        """
        are we running in a public Continuous Integration Environment?
        """
        publicCI = getpass.getuser() in ["travis", "runner"]
        jenkins = "JENKINS_HOME" in os.environ
        return publicCI or jenkins

    @staticmethod
    def isUser(name: str):
        """Checks if the system has the given name"""
        return getpass.getuser() == name


class Profiler:
    """
    simple profiler
    """

    def __init__(self, msg, profile=True):
        """
        construct me with the given msg and profile active flag

        Args:
            msg(str): the message to show if profiling is active
            profile(bool): True if messages should be shown
        """
        self.msg = msg
        self.profile = profile
        self.starttime = time.time()
        if profile:
            print(f"Starting {msg} ...")

    def time(self, extraMsg=""):
        """
        time the action and print if profile is active
        """
        elapsed = time.time() - self.starttime
        if self.profile:
            print(f"{self.msg}{extraMsg} took {elapsed:5.1f} s")
        return elapsed



'''
Created on 2022-11-16

@author: wf
'''
from tests.basetest import Basetest
from skg.scholar import Scholar
from skg.kg import SKG_Def

class TestScholar(Basetest):
    """
    test concerning the scholar/author concept
    """
    
    def test_scholar_by_id(self):
        """
        test searching a scholar/author by ORCID
        """
        id_examples=[
            {
                "id_name": "orcid",
                "id_value": "0000-0003-1279-3709"
            },
            {
                "id_name": "dblpId",
                "id_value": "b/TimBernersLee"
            },
            {
                "id_name": "wikiDataId",
                "id_value": "Q80"
            },           
        ]
        debug=self.debug
        debug=True
        skg_def=SKG_Def()
        for id_example in id_examples:
            id_name=id_example["id_name"]
            id_value=id_example["id_value"]
            scholars=Scholar.from_wikidata_via_id(skg_def.concepts["Scholar"],id_name,id_value)
            if debug:
                for scholar in scholars:
                    print(scholar)
            self.assertEqual(1,len(scholars))
            scholar=scholars[0]
            self.assertEqual("Tim Berners-Lee",scholar.label)
            self.assertEqual("https://scholia.toolforge.org/author/Q80",scholar.scholia_url())



'''
Created on 2022-11-16

@author: wf
'''
import skg.graph

class Event(skg.graph.Node):
    """
    an instance of a scientific event
    """
    
    @classmethod
    def getSamples(cls):
        samples=[
            {
                "wikiDataId":"Q112055391",
                "title": "The Third Wikidata Workshop",
                "location": "Hangzhou",
                "point_in_time": "2022-10-24",
                "official_website": "https://wikidataworkshop.github.io/2022/"
            }
        ]
        return samples
                



'''
Created on 2022-11-16

@author: wf
'''
import skg.graph

class Paper(skg.graph.Node):
    '''
    a scientific paper
    '''

    @classmethod
    def getSamples(cls):
        samples=[
            { 
                "wikiDataId": "Q55693406",
                "title":"Designing the web for an open society",
                "DOI": "10.1145/1963405.1963408",
                "DBLP_publication_ID": "conf/www/Berners-Lee11", 
                "publication_date": 2011,
            }
        ]
        return samples
        
    def __init__(self):
        '''
        Constructor
        '''
        


'''
Created on 2022-11-16

@author: wf
'''
from tests.basetest import Basetest
from skg.wdsearch import WikidataSearch
from skg.wikidata import Wikidata
import json
from lodstorage.lod import LOD

class TestWikidataSearch(Basetest):
    """
    test the wikidata search
    """
    
    def test_wikidata_search(self):
        """
        test the wikidata serarch API
        """
        search="Tim Berners-Lee"
        wd=WikidataSearch()
        search_options=wd.searchOptions(search)
        self.assertTrue(len(search_options)>0)
        qid,_itemLabel,_desc=search_options[0]
        self.assertEqual("Q80",qid)
        pass
    
    def test_item_instanceof(self):
        """
        test getting the classes Q-Identifiers of a given instance 
        """
        wd=Wikidata()
        qids=["Q80","Q937","Q112055391"]
        expected=["Q5","Q5","Q52260246"]
        classes_map=wd.getClassQids(qids)
        if self.debug:
            print(json.dumps(classes_map,indent=2))
        for i,qid in enumerate(qids):
            classes=classes_map[qid]
            classes_by_class_qid,_dup=LOD.getLookup(classes, "class_qid")
            self.assertTrue(expected[i] in classes_by_class_qid)
            pass



'''
Created on 2022-04-01

@author: wf
'''
import skg

class Version(object):
    '''
    Version handling for pysotsog
    '''
    name="pysotsog"
    description='''python Library for Scholars to achieve "Standing on the shoulders of giants"'''
    version=skg.__version__
    date = '2022-11-16'
    updated = '2022-11-16'

        


'''
Created on 2022-11-16

@author: wf
'''
from tests.basetest import Basetest
from skg.sotsog import SotSog
class TestSotsog(Basetest):
    """
    test concerning the SotSog main class
    """

    def test_sotsog_search(self):
        """
        test searches
        """
        sotsog=SotSog()
        search_examples=[
            {
                "search": ["Wikidata Workshop 2022"],
            }, 
            {
                "search": ["Albert","Einstein"],
            },
            {
                "search": ["Designing the web for an open society"]
            },
        ]
        for _i,search_example in enumerate(search_examples):
            search=search_example["search"]
            items=sotsog.search(search,show=False, open_browser=False)        
            for item in items:
                print(item)



'''
Created on 2022-11-16

@author: wf
'''

import os
import webbrowser
import sys
import traceback
from skg.version import Version
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
from skg.wdsearch import WikidataSearch
from skg.wikidata import Wikidata
from skg.kg import SKG_Def
from skg.graph import Node

class SotSog():
    """
    Standing on the shoulders of giants
    """
    
    def __init__(self,debug:bool=False):
        """
        constructor
        
        Args:
            debug(bool): if True debugging should be switched on
        """
        self.debug=debug
        Node.debug=debug
        self.wikipedia_url="https://en.wikipedia.org/wiki/Standing_on_the_shoulders_of_giants"
        self.skg_def=SKG_Def()
        self.scholar_concept=self.skg_def.concepts["Scholar"]
    
    def search(self,search_list,lang='en',show:bool=True,open_browser:bool=False)->list:
        """
        search with the given search list
        """
        search_term=' '.join(search_list)
        wd=Wikidata(debug=self.debug)
        wds=WikidataSearch(language=lang,debug=self.debug)
        search_options=wds.searchOptions(search_term)
        items=[]
        qids=[]
        for qid,itemLabel,desc in search_options:
            qids.append(qid)
        class_map=wd.getClassQids(qids)
        for qid,itemLabel,desc in search_options:
            if qid in class_map:
                class_rows=class_map[qid]
                for class_row in class_rows:
                    class_qid=class_row["class_qid"]
                    concept=self.skg_def.conceptForQid(class_qid)
                    if concept is not None:
                        wd_items=concept.cls.from_wikidata_via_id(concept,"wikiDataId", qid, lang=lang)
                        if len(wd_items)>0:
                            item=wd_items[0]
                            items.append(item)
                            if show:
                                print(f"{itemLabel}({qid}):{desc}✅")
                                print(item)
                            if open_browser:
                                scholia_url=item.scholia_url()
                                print(f"opening {scholia_url} in browser")
                                webbrowser.open(scholia_url)
        return items

__version__ = Version.version
__date__ = Version.date
__updated__ = Version.updated


def main(argv=None): # IGNORE:C0111
    '''main program.'''

    if argv is None:
        argv=sys.argv[1:]
    
    program_name = os.path.basename(__file__)
    program_shortdesc = Version.description
    
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)

    user_name="Wolfgang Fahl"
    program_license = '''%s

  Created by %s on %s.
  Copyright 2022 Wolfgang Fahl. All rights reserved.

  Licensed under the Apache License 2.0
  http://www.apache.org/licenses/LICENSE-2.0

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
''' % (program_shortdesc, user_name,str(__date__))
    try:
        # Setup argument parser
        parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
        parser.add_argument('search', action='store', nargs='*', help="search terms")
        parser.add_argument("-d", "--debug", dest="debug", action="store_true", help="show debug info")
        parser.add_argument("-l", "--lang",help="language code to use",default="en")
        parser.add_argument('-V', '--version', action='version', version=program_version_message)
        args = parser.parse_args(argv)
        if len(argv) < 2:
            parser.print_usage()
            sys.exit(1)
        sotsog=SotSog(debug=args.debug)
        sotsog.search(args.search,lang=args.lang,open_browser=True)
        pass
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 1
    except Exception as e:
        if DEBUG:
            raise(e)
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        if args.debug:
            print(traceback.format_exc())
        return 2       
        
DEBUG = 1
if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(main())



'''
Created on 2022-11-16

@author: wf
'''
from skg.scholar import Scholar
from skg.paper import Paper
from skg.event import Event
from skg.graph import Concept


class SKG_Def:
    """
    scholary knowledge graph
    """
    
    def __init__(self):
        """
        constructor
        """
        self.concepts={
            "Scholar": Concept(name="Scholar",cls=Scholar),
            "Paper": Concept(name="Paper",cls=Paper),
            "Event": Concept(name="Event",cls=Event)
        }
        self.concepts["Scholar"].map_wikidata("Q5","author",[
            ("name","label"),
            ("dblpId","P2456"),("gndId","P227"),
            ("linkedInId","P6634"),
            ("homepage","P856"),
            ("googleScholarUser","P1960"),("orcid","P496"),
            ("givenName","P735"),
            ("familyName","P734")
        ])
        self.concepts["Paper"].map_wikidata("Q13442814","work",[
            ("title","label"),
            ("DOI","P356"),
            ("DBLP_publication_ID","P8978"),
            ("publication_date","P577")
        ])
        # scientific event
        self.concepts["Event"].map_wikidata("Q52260246","event",[
            ("title","P1476"),
            ("location","P276"),
            ("point_in_time","P585"),
            ("official_website","P856")
        ])
                      
        self.concepts_by_qid={}
        for concept in self.concepts.values():
            if concept.wd_class in self.concepts_by_qid:
                raise Exception(f"duplicate wd_class definition: {concept.wd_class}")
            self.concepts_by_qid[concept.wd_class]=concept
        
    def conceptForQid(self,qid:str)->Concept:
        """
        get the concept for the given wikidata Q Identifieer
        
        Args:
            qid(str): get the concept for the given Qid
            
        Return:
            Concept: or None if none is found
        """
        concept=self.concepts_by_qid.get(qid,None)
        return concept
        


'''
Created on 2022-11-16

@author: wf
'''
from lodstorage.sparql import SPARQL
from lodstorage.lod import LOD
class Wikidata:
    """
    Wikidata access wrapper
    """
    def __init__(self,endpoint:str="https://query.wikidata.org/sparql",debug:bool=False):
        """
        constructor
        """
        self.endpoint=endpoint
        self.sparql = SPARQL(endpoint)
        self.debug=debug
        
        
    def getClassQids(self,qids:list)->dict:
        """
        get the Wikidata Q-Identifiers 
        for the given wikidata ids
        
        Args:
            qids(list): the list of id
        """
        sparql_query=f"""# get the instanceof values for a given entity
SELECT ?item ?itemLabel ?qid ?class_qid ?class ?classLabel
WHERE 
{{
  VALUES ?item {{
"""
        for qid in qids:
            if not qid.startswith("http:"):
                wd_url=f"http://www.wikidata.org/entity/{qid}"
            else:
                wd_url=qid
            sparql_query+=f"    <{wd_url}>\n"
        sparql_query+=f"""}}
  ?item wdt:P31/wdt:P279* ?class.
  ?item rdfs:label ?itemLabel
  FILTER(LANG(?itemLabel)="en")
  ?class rdfs:label ?classLabel
  FILTER(LANG(?classLabel)="en")
  BIND(REPLACE(STR(?class),"http://www.wikidata.org/entity/","") AS ?class_qid)
  BIND(REPLACE(STR(?item),"http://www.wikidata.org/entity/","") AS ?qid)
}}"""
        if self.debug:
            print(sparql_query)
        class_rows=self.sparql.queryAsListOfDicts(sparql_query)
        class_map=LOD.getLookup(class_rows, "qid", withDuplicates=True)
        return class_map



'''
Created on 2022-11-16

@author: wf
'''
from skg.wikidata import Wikidata

class Concept:
    """
    an Entity
    """
    def __init__(self,name:str,cls):
        """
        constructor
        
        Args:
            name(str): the name of the node
            cls: a class
        """
        self.name=name
        self.props={}
        self.cls=cls
        # @TODO check that getSamples exists
        for sample in cls.getSamples():
            for key in sample.keys():
                if not key in self.props:
                    self.props[key]=Property(key)
                    
    def map_wikidata(self,wd_class:str,scholia_suffix,map_list:list):
        """
        map wikidata entries
        
        Args:
            wd_class(str): the main wikidata base class
            scholia_suffix(str): the scholia suffix
        """
        self.wd_class=wd_class
        self.scholia_suffix=scholia_suffix
        for prop_name,wd_prop in map_list:
            if prop_name in self.props:
                self.props[prop_name].wd_prop=wd_prop
            
class Property:
    """
    a Property
    """
    def __init__(self,name):
        self.name=name
    
class Node:
    """
    a Node in the scholary knowledge graph
    """
    debug=False
    
    def __init__(self):
        """
        constructor
        """
    
    def __str__(self):
        """
        return a text representation of me
        """
        text=f"{self.concept.name} {self.label}:"
        delim="\n  "
        for prop in self.concept.props.values():
            if hasattr(self, prop.name):
                text+=f"{delim}{prop.name}={getattr(self,prop.name)}"
        return text
        
    def from_dict(self,concept,record:str):
        """
        get my values from the given record
        """
        self.concept=concept
        self.label=record[concept.name]
        for key in concept.props.keys():
            if key in record:
                setattr(self, key, record[key])
                
    def scholia_url(self):
        """
        get my scholia url
        """
        prefix=f"https://scholia.toolforge.org/{self.concept.scholia_suffix}"
        wd_url=getattr(self, "wikiDataId",None)
        if wd_url is None:
            return prefix
        else:
            qid=wd_url.replace("http://www.wikidata.org/entity/","")
            return f"{prefix}/{qid}"
        
    @classmethod
    def from_wikidata_via_id(cls,concept,id_name:str,id_value:str,lang:str="en"):
        wikidata=Wikidata()
        if id_name=="wikiDataId":
            value_clause=f"<http://www.wikidata.org/entity/{id_value}>"
        else:
            value_clause=f'''"{id_value}"'''
        sparql_query=f"""# Query for {concept.name} details via ID {id_name} value {id_value}
PREFIX wd: <http://www.wikidata.org/entity/>
PREFIX wdt: <http://www.wikidata.org/prop/direct/>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT ?{concept.name} ?qId"""
        for prop in concept.props.values():
            sparql_query+=f" ?{prop.name}"
        sparql_query+=f"""
WHERE {{
  VALUES ?{id_name} {{
    {value_clause}
  }}
  ?wikiDataId wdt:P31/wdt:P279* wd:{concept.wd_class}.
  ?wikiDataId rdfs:label ?{concept.name} .
  FILTER(LANG(?{concept.name})="{lang}").
"""
        for prop in concept.props.values():
            if prop.name=="wikiDataId":
                continue
            if not hasattr(prop, "wd_prop"):
                raise Exception(f"Property {prop.name} has no wikidata mapping")
            clause=f"?wikiDataId wdt:{prop.wd_prop} ?{prop.name}."
            if prop.name!=id_name:
                clause=f"OPTIONAL {{ {clause} }}"
            sparql_query+="\n  "+clause
        sparql_query+="\n}"
        if Node.debug:
            print(sparql_query)
        records=wikidata.sparql.queryAsListOfDicts(sparql_query)
        instances=[]
        for record in records:
            instance=cls()
            instance.from_dict(concept,record)
            instances.append(instance)
        return instances


'''
Created on 2022-11-16

@author: wf
'''
from tests.basetest import Basetest
from skg.kg import SKG_Def
class TestSkg(Basetest):
    """
    test scholarly knowledge graph
    """
    
    def test_skg(self):
        """
        test the scholarly knowledge graph definition
        """
        skg_def=SKG_Def()
        concept=skg_def.conceptForQid("Q5")
        self.assertEqual("Scholar",concept.name)


'''
Created on 24.07.2022

@author: wf
'''
import json
import os
import urllib.request
import urllib.parse

class WikidataSearch(object):
    '''
    Wikidata Search
    '''

    def __init__(self,language='en',timeout=2.0,debug:bool=False):
        '''
        Constructor
        
        Args:
            language(str): the language to use e.g. en/fr
            timeout(float): maximum time to wait for result
            debug(bool): if True debug details should be shown
        '''
        self.language=language
        self.timeout=timeout
        self.debug=debug
        
    def searchOptions(self,searchFor:str,limit:int=9)->list:
        '''
        search and return a list of qid,itemLabel description tuples
        
        Args:
            searchFor(str): the string to search for
            limit(int): the maximum amount of results to search for
        '''
        options=[]
        srlist=self.search(searchFor, limit)
        if srlist is not None:
            for sr in srlist:
                qid=sr["id"]
                itemLabel=sr["label"]
                desc=""
                if "display" in sr:
                    display=sr["display"]
                    if "description" in display:
                        desc=display["description"]["value"]
                options.append((qid,itemLabel,desc,))
        return options

    def search(self,searchFor:str,limit:int=9):
        '''
        
        Args:
            searchFor(str): the string to search for
            limit(int): the maximum amount of results to search for
        '''
        try:
            apiurl=f"https://www.wikidata.org/w/api.php?action=wbsearchentities&language={self.language}&format=json&limit={limit}&search="
            searchEncoded=urllib.parse.quote_plus(searchFor)
            apisearch=apiurl+searchEncoded
            with urllib.request.urlopen(apisearch,timeout=self.timeout) as url:
                searchResult = json.loads(url.read().decode())
            return searchResult["search"]
        except Exception as _error:
            return None
        
    def getProperties(self):
        '''
        get the Wikidata Properties 
        '''
        scriptdir=os.path.dirname(__file__)
        jsonPath=f"{scriptdir}/resources/wdprops.json"
        with open(jsonPath) as jsonFile:
            props=json.load(jsonFile)
        return props